# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/pote',
                'doc_host': 'https://civvic.github.io',
                'git_url': 'https://github.com/civvic/pote',
                'lib_path': 'pote'},
  'syms': { 'pote.all': {},
            'pote.basic': { 'pote.basic.AD': ('basic.html#ad', 'pote/basic.py'),
                            'pote.basic.AD.__dir__': ('basic.html#ad.__dir__', 'pote/basic.py'),
                            'pote.basic.AD.__getattr__': ('basic.html#ad.__getattr__', 'pote/basic.py'),
                            'pote.basic.AD.__setattr__': ('basic.html#ad.__setattr__', 'pote/basic.py'),
                            'pote.basic.AD._repr_markdown_': ('basic.html#ad._repr_markdown_', 'pote/basic.py'),
                            'pote.basic.AD.copy': ('basic.html#ad.copy', 'pote/basic.py'),
                            'pote.basic.Empty': ('basic.html#empty', 'pote/basic.py'),
                            'pote.basic.EmptyT': ('basic.html#emptyt', 'pote/basic.py'),
                            'pote.basic.Kounter': ('basic.html#kounter', 'pote/basic.py'),
                            'pote.basic.Kounter.__call__': ('basic.html#kounter.__call__', 'pote/basic.py'),
                            'pote.basic.Kounter.__init__': ('basic.html#kounter.__init__', 'pote/basic.py'),
                            'pote.basic.Runner': ('basic.html#runner', 'pote/basic.py'),
                            'pote.basic.WithCounterMeta': ('basic.html#withcountermeta', 'pote/basic.py'),
                            'pote.basic.WithCounterMeta.__call__': ('basic.html#withcountermeta.__call__', 'pote/basic.py'),
                            'pote.basic.WithCounterMeta.__new__': ('basic.html#withcountermeta.__new__', 'pote/basic.py'),
                            'pote.basic._get_globals': ('basic.html#_get_globals', 'pote/basic.py'),
                            'pote.basic._vals_at': ('basic.html#_vals_at', 'pote/basic.py'),
                            'pote.basic._vals_atpath': ('basic.html#_vals_atpath', 'pote/basic.py'),
                            'pote.basic.bundle_path': ('basic.html#bundle_path', 'pote/basic.py'),
                            'pote.basic.deep_in': ('basic.html#deep_in', 'pote/basic.py'),
                            'pote.basic.flatten': ('basic.html#flatten', 'pote/basic.py'),
                            'pote.basic.gets': ('basic.html#gets', 'pote/basic.py'),
                            'pote.basic.has_key': ('basic.html#has_key', 'pote/basic.py'),
                            'pote.basic.has_path': ('basic.html#has_path', 'pote/basic.py'),
                            'pote.basic.id_gen': ('basic.html#id_gen', 'pote/basic.py'),
                            'pote.basic.is_empty': ('basic.html#is_empty', 'pote/basic.py'),
                            'pote.basic.is_listy': ('basic.html#is_listy', 'pote/basic.py'),
                            'pote.basic.is_listy_type': ('basic.html#is_listy_type', 'pote/basic.py'),
                            'pote.basic.pops_': ('basic.html#pops_', 'pote/basic.py'),
                            'pote.basic.pops_values_': ('basic.html#pops_values_', 'pote/basic.py'),
                            'pote.basic.setattrs': ('basic.html#setattrs', 'pote/basic.py'),
                            'pote.basic.shorten': ('basic.html#shorten', 'pote/basic.py'),
                            'pote.basic.shortens': ('basic.html#shortens', 'pote/basic.py'),
                            'pote.basic.simple_id': ('basic.html#simple_id', 'pote/basic.py'),
                            'pote.basic.update_': ('basic.html#update_', 'pote/basic.py'),
                            'pote.basic.val_at': ('basic.html#val_at', 'pote/basic.py'),
                            'pote.basic.val_atpath': ('basic.html#val_atpath', 'pote/basic.py'),
                            'pote.basic.vals_at': ('basic.html#vals_at', 'pote/basic.py'),
                            'pote.basic.vals_atpath': ('basic.html#vals_atpath', 'pote/basic.py')},
            'pote.callback': { 'pote.callback.Callback': ('callback.html#callback', 'pote/callback.py'),
                               'pote.callback.CollBack': ('callback.html#collback', 'pote/callback.py'),
                               'pote.callback.CollBack.__init__': ('callback.html#collback.__init__', 'pote/callback.py'),
                               'pote.callback.CollBack.__iter__': ('callback.html#collback.__iter__', 'pote/callback.py'),
                               'pote.callback.CollBack.__repr__': ('callback.html#collback.__repr__', 'pote/callback.py'),
                               'pote.callback.CollBack._interrupt': ('callback.html#collback._interrupt', 'pote/callback.py'),
                               'pote.callback.CollBack._start': ('callback.html#collback._start', 'pote/callback.py'),
                               'pote.callback.CollBack._stop': ('callback.html#collback._stop', 'pote/callback.py'),
                               'pote.callback.CollBack.progress': ('callback.html#collback.progress', 'pote/callback.py'),
                               'pote.callback.CollBack.state': ('callback.html#collback.state', 'pote/callback.py'),
                               'pote.callback.CollBack.trackback': ('callback.html#collback.trackback', 'pote/callback.py'),
                               'pote.callback.EchoCB': ('callback.html#echocb', 'pote/callback.py'),
                               'pote.callback.EchoCB.__getattr__': ('callback.html#echocb.__getattr__', 'pote/callback.py'),
                               'pote.callback.EchoCB.echo': ('callback.html#echocb.echo', 'pote/callback.py'),
                               'pote.callback.FuncCB': ('callback.html#funccb', 'pote/callback.py'),
                               'pote.callback.FuncCB.__init__': ('callback.html#funccb.__init__', 'pote/callback.py'),
                               'pote.callback.HasCallbacks': ('callback.html#hascallbacks', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.__getattr__': ('callback.html#hascallbacks.__getattr__', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.__init__': ('callback.html#hascallbacks.__init__', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.__new__': ('callback.html#hascallbacks.__new__', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.callback': ('callback.html#hascallbacks.callback', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.this_cbs': ('callback.html#hascallbacks.this_cbs', 'pote/callback.py'),
                               'pote.callback.HasCallbacks.with_cbs': ('callback.html#hascallbacks.with_cbs', 'pote/callback.py'),
                               'pote.callback.PassCB': ('callback.html#passcb', 'pote/callback.py'),
                               'pote.callback.PassCB.__getattr__': ('callback.html#passcb.__getattr__', 'pote/callback.py'),
                               'pote.callback.PassCB.noop': ('callback.html#passcb.noop', 'pote/callback.py'),
                               'pote.callback._get_total': ('callback.html#_get_total', 'pote/callback.py'),
                               'pote.callback.process_': ('callback.html#process_', 'pote/callback.py'),
                               'pote.callback.run_cbs': ('callback.html#run_cbs', 'pote/callback.py'),
                               'pote.callback.trackback': ('callback.html#trackback', 'pote/callback.py'),
                               'pote.callback.with_cbs': ('callback.html#with_cbs', 'pote/callback.py'),
                               'pote.callback.with_cbs.__call__': ('callback.html#with_cbs.__call__', 'pote/callback.py'),
                               'pote.callback.with_cbs.__init__': ('callback.html#with_cbs.__init__', 'pote/callback.py')},
            'pote.common': {},
            'pote.config': { 'pote.config.Config': ('config.html#config', 'pote/config.py'),
                             'pote.config.Config.__call__': ('config.html#config.__call__', 'pote/config.py'),
                             'pote.config.Config.__repr__': ('config.html#config.__repr__', 'pote/config.py'),
                             'pote.config.Config.as_dict': ('config.html#config.as_dict', 'pote/config.py'),
                             'pote.config.Config.update': ('config.html#config.update', 'pote/config.py')},
            'pote.display': { 'pote.display.GlobalCSS': ('display.html#globalcss', 'pote/display.py'),
                              'pote.display.GlobalCSS.__init__': ('display.html#globalcss.__init__', 'pote/display.py'),
                              'pote.display.GlobalCSS._html_widget': ('display.html#globalcss._html_widget', 'pote/display.py'),
                              'pote.display.GlobalCSS._ipython_display_': ('display.html#globalcss._ipython_display_', 'pote/display.py'),
                              'pote.display.GlobalCSS.add': ('display.html#globalcss.add', 'pote/display.py'),
                              'pote.display.GlobalCSS.css': ('display.html#globalcss.css', 'pote/display.py'),
                              'pote.display.GlobalCSS.display': ('display.html#globalcss.display', 'pote/display.py'),
                              'pote.display.GlobalCSS.has_style': ('display.html#globalcss.has_style', 'pote/display.py'),
                              'pote.display.GlobalCSS.update': ('display.html#globalcss.update', 'pote/display.py'),
                              'pote.display.RenderJSON': ('display.html#renderjson', 'pote/display.py'),
                              'pote.display.RenderJSON.__init__': ('display.html#renderjson.__init__', 'pote/display.py'),
                              'pote.display.RenderJSON._ipython_display_': ('display.html#renderjson._ipython_display_', 'pote/display.py'),
                              'pote.display.RenderJSON.display': ('display.html#renderjson.display', 'pote/display.py'),
                              'pote.display.cssmap': ('display.html#cssmap', 'pote/display.py'),
                              'pote.display.pretty_repr': ('display.html#pretty_repr', 'pote/display.py'),
                              'pote.display.rich_display': ('display.html#rich_display', 'pote/display.py')},
            'pote.project': { 'pote.project.get_console': ('project.html#get_console', 'pote/project.py'),
                              'pote.project.get_theme': ('project.html#get_theme', 'pote/project.py'),
                              'pote.project.setup_console': ('project.html#setup_console', 'pote/project.py')},
            'pote.test': { 'pote.test.test_afail': ('test.html#test_afail', 'pote/test.py'),
                           'pote.test.test_is_not': ('test.html#test_is_not', 'pote/test.py'),
                           'pote.test.test_raises': ('test.html#test_raises', 'pote/test.py')},
            'pote.widget': { 'pote.widget.Clickable': ('widgets.html#clickable', 'pote/widget.py'),
                             'pote.widget.Clickable.__init__': ('widgets.html#clickable.__init__', 'pote/widget.py'),
                             'pote.widget.cleanupwidgets': ('widgets.html#cleanupwidgets', 'pote/widget.py'),
                             'pote.widget.close_widget': ('widgets.html#close_widget', 'pote/widget.py'),
                             'pote.widget.get_user_input': ('widgets.html#get_user_input', 'pote/widget.py'),
                             'pote.widget.wait_for_change': ('widgets.html#wait_for_change', 'pote/widget.py'),
                             'pote.widget.yield_for_change': ('widgets.html#yield_for_change', 'pote/widget.py')}}}



"""Utilities for ipywidgets: cleanup, async interaction, and blocking input"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/20_widgets.ipynb.

# %% ../nbs/20_widgets.ipynb 1
from __future__ import annotations

# %% auto 0
__all__ = ['close_widget', 'cleanupwidgets', 'Clickable', 'wait_for_change', 'yield_for_change', 'get_user_input']

# %% ../nbs/20_widgets.ipynb
import asyncio
import time
from functools import wraps

import ipywidgets as W
import traitlets as T
from IPython.display import display
from IPython.display import Markdown
from jupyter_ui_poll import ui_events


# %% ../nbs/20_widgets.ipynb
from .basic import _get_globals


# %% ../nbs/20_widgets.ipynb
def close_widget(w: W.Widget, all: bool=True):
    if all:
        for k in w.keys:  # type: ignore
            if isinstance((v := getattr(w, k)), W.Widget):
                close_widget(v)
    w.close()

# %% ../nbs/20_widgets.ipynb
def cleanupwidgets(*ws, mod: str|None=None, clear=True, all=True):
    from IPython.display import clear_output
    glb = _get_globals(mod or __name__)
    for w in ws:
        _w = glb.get(w) if isinstance(w, str) else w
        if _w:
            try: close_widget(_w, all=all)
            except Exception as e: pass
    if clear: clear_output(wait=False)

# %% ../nbs/20_widgets.ipynb
class Clickable(W.Button, W.ValueWidget):
    "Button with value."
    value = T.Int(0)
    def __init__(self, description='', **kwargs):
        super().__init__(description=description, **kwargs)
        self.on_click(lambda b: b.set_trait('value', b.value + 1))
        if not description:
            T.dlink((self, 'value'), (self, 'description'), lambda x: f'{x}')

# %% ../nbs/20_widgets.ipynb
def wait_for_change(widget:W.Widget, value:str):
    future = asyncio.Future()
    def getvalue(change):
        future.set_result(change.new)
        widget.unobserve(getvalue, value)
    widget.observe(getvalue, value)
    return future

# %% ../nbs/20_widgets.ipynb
def yield_for_change(widget, attribute):
    """Pause a generator to wait for a widget change event.
        
    This is a decorator for a generator function which pauses the generator on yield
    until the given widget attribute changes. The new value of the attribute is
    sent to the generator and is the value of the yield.
    """
    def f(iterator):
        @wraps(iterator)
        def inner():
            i = iterator()
            def next_i(change):
                try: i.send(change.new)
                except StopIteration as e: widget.unobserve(next_i, attribute)
            widget.observe(next_i, attribute)
            # start the generator
            next(i)
        return inner
    return f

# %% ../nbs/20_widgets.ipynb
def get_user_input(
        prompt='', placeholder='Write something. Enter to submit', timeout=10., widget=None, value=None):
    if widget is None:
        layout = {'description_width':'auto', 'width':'80%'}
        b = W.HBox([
            W.HTML(f"<b>{prompt}</b>" if prompt else ''),
            w := W.Text(value=value,placeholder=placeholder, continuous_update=False, layout=layout)
        ])
        dh = display(b, display_id=True); time.sleep(0.05)
    else: w, dh = widget, None
    w.focus()
    answer = v = w.value
    start_time = time.time()
    with ui_events() as ui_poll:
        while answer == v:
            if (time.time() - start_time) > timeout: break
            ui_poll(10)
            time.sleep(0.1)  # Simulate async processing
            answer = w.value
    if dh: b.close(); dh.update(Markdown(f"**{prompt}** {answer}" if prompt else answer))
    return answer.lower()



"""Configuration container with temporary context manager for scoped changes"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/15_config.ipynb.

# %% ../nbs/15_config.ipynb 1
from __future__ import annotations

# %% auto 0
__all__ = ['Config']

# %% ../nbs/15_config.ipynb
import inspect
from contextlib import contextmanager

import fastcore.all as FC


# %% ../nbs/15_config.ipynb
class Config:
    "Store of arbitrary values"
    def __repr__(self): return str(self.as_dict())
    def update(self, **kwargs):
        ks = set(FC.flatten(map(inspect.get_annotations, type(self).mro())))
        for k,v in kwargs.items():
            if k in ks:
                try: setattr(self, k, v)
                except AttributeError: pass
        return self
    def as_dict(self): return {k:getattr(self, k) for k in 
        FC.flatten(map(inspect.get_annotations, type(self).mro())) if hasattr(self, k)}
    @contextmanager
    def __call__(self, **kwargs):
        vv = vars(self).copy()
        old = {k:vv[k] for k in kwargs if k in vv}
        for k,v in kwargs.items(): setattr(self, k, v)
        yield
        for k in [k for k in vars(self).keys() if k not in vv]: delattr(self, k)
        for k,v in old.items(): setattr(self, k, v)



"""Extensible callback system for augmenting objects and tracking iterations"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_callback.ipynb.

# %% ../nbs/10_callback.ipynb 1
from __future__ import annotations

# %% auto 0
__all__ = ['Callback', 'run_cbs', 'EchoCB', 'FuncCB', 'PassCB', 'HasCallbacks', 'with_cbs', 'CollBack', 'trackback', 'process_']

# %% ../nbs/10_callback.ipynb
import collections
import time
from contextlib import contextmanager
from functools import partial
from operator import attrgetter
from operator import length_hint
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import Sequence
from typing import Type
from typing import TypeVar

import fastcore.all as FC


# %% ../nbs/10_callback.ipynb
from .basic import AD
from .basic import empty
from .basic import EmptyT
from .basic import update_
from .common import Runner
from .common import setattrs


# %% ../nbs/10_callback.ipynb
class Callback:
    "Base class of callbacks."
    order = 0
    cbs: Sequence[Callback]  # if present, run callbacks in the list after running this callback

# %% ../nbs/10_callback.ipynb
def run_cbs(cbs: Iterable[Callback] | FC.L, method_nm:str, ctx=None, *args, **kwargs):
    "Run `method_nm(ctx, ...)` of each callback in `cbs` in order."
    for cb in sorted(cbs, key=attrgetter('order')):
        if f := getattr(cb, method_nm, None): Runner(f)(ctx, *args, **kwargs)
        if nested := getattr(cb, 'cbs', None): 
            run_cbs(filter(lambda x: x != cb, nested), method_nm, ctx, *args, **kwargs)

# %% ../nbs/10_callback.ipynb
class EchoCB(Callback):
    "Print the arguments."
    cbs=()
    def echo(self, ctx, *args, **kwargs): print(ctx, args, kwargs)
    def __getattr__(self, name): return partial(self.echo, name)

# %% ../nbs/10_callback.ipynb
class FuncCB(Callback):
    "Store functions that can be called as callbacks."
    def __init__(self, **kwargs): setattrs(self, kwargs)

# %% ../nbs/10_callback.ipynb
class PassCB(Callback):
    "A callback that does nothing."
    def noop(self, *args, **kwargs): pass
    def __getattr__(self, name): return self.noop

# %% ../nbs/10_callback.ipynb
class HasCallbacks:
    "Base for classes that can be augmented with callbacks."
    cbs: list[Callback]
    cbs_names: tuple[str,...] = ()  # this callbacks becomes self methods if a subclass overrides them

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        self.cbs = []#PassCB()]
        return self

    def __init__(self, cbs:Sequence[Callback]=()): self.cbs = list(cbs)

    def with_cbs(self, cbs:Sequence[Callback], extend=False):
        if extend: self.cbs.extend(cbs)
        else: self.cbs = list(cbs)
        return self
    
    @contextmanager
    def this_cbs(self, cbs:Sequence[Callback]):
        "Use temporary `cbs` in `with` block."
        for cb in cbs: self.cbs.append(cb)
        try:
            yield        
        finally:
            for cb in cbs: self.cbs.remove(cb)

    def __getattr__(self, name):
        "Run `name` as a callback if it's in `self.cbs_names`."
        if name in self.cbs_names: return partial(self.callback, name)
        raise AttributeError(name)

    def callback(self, method_nm, *args, **kwargs): run_cbs(self.cbs, method_nm, self, *args, **kwargs)

# %% ../nbs/10_callback.ipynb
class with_cbs:
    "Decorator to add callbacks to a method."
    def __init__(self, nm:str|None=None): self.nm = nm
    def __call__(self, f):
        def _f(o, *args, **kwargs):
            nm = self.nm or f.__name__
            try:
                o.callback(f'before_{nm}')
                f(o, *args, **kwargs)
                o.callback(f'after_{nm}')
            except globals()[f'Cancel{nm.title()}Exception']: pass
            finally: o.callback(f'cleanup_{nm}')
        return _f

# %% ../nbs/10_callback.ipynb
def _get_total(total: int|None|Type[EmptyT], source) -> int|None:
    if total is empty:
        try: return len(source)
        except: return length_hint(source) or None
    if total is not None and (not isinstance(total, int) or total < 0): total = None
    return total

# %% ../nbs/10_callback.ipynb
class CollBack(HasCallbacks):
    "Track iterables and extend them with callbacks."

    cbs_names = ('before_iter', 'after_iter', 'on_iter', 'on_interrupt')

    n: int|None = None
    elapsed_time: float|None = None
    active, interrupted = False, False
    item: Any = empty
    def __init__(self, 
            source: Iterable[Any] = (), 
            total: int|None|Type[EmptyT] = empty, 
            context: Any = empty,
            **kwargs):
        self._source, self._total = source, total
        self.total: int|None = _get_total(self._total, self._source)
        self.context = context
        super().__init__(**kwargs)

    @property
    def progress(self):
        if self.total:
            return None if self.n is None else min(1., round((self.n+1)/float(self.total), 4))
        return None

    @property
    def state(self):
        return AD(update_(item=self.item, n=self.n, total=self.total, progress=self.progress, 
            context=self.context, 
            elapsed_time=self.elapsed_time, interrupted=self.interrupted or empty, empty_value=empty))

    def __repr__(self): return f'{self.__class__.__name__}#{self._source}, total={self._total}'

    def _start(self):
        self.total = _get_total(self._total, self._source)
        self.active, self.elapsed_time, self.n = True, None, None
        run_cbs(self.cbs, 'before_iter', self.state)

    def _stop(self):
        if self.total is None and self.n is not None: self.total = self.n + 1
        self.active, self.item = False, empty
        run_cbs(self.cbs, 'after_iter', self.state)
    
    def _interrupt(self): self.interrupted = True; run_cbs(self.cbs, 'on_interrupt', self.state)

    def __iter__(self) -> Iterator[Any]:
        if self._source is None: return
        try:
            start_time = time.time()
            self._start()
            for self.n, self.item in enumerate(self._source):
                if self.total is not None and self.n >= self.total: break
                yield self.item
                self.elapsed_time = time.time() - start_time
                run_cbs(self.cbs, 'on_iter', self.state, self.item)
                if self.total is not None and self.n >= self.total-1: break
        except Exception as e: self._interrupt(); raise e
        finally:
            self._stop()

    def trackback(self, cbs:Sequence[Callback]=()) -> Iterator[tuple[AD[Any], Any]]:
        with self.this_cbs(cbs):
            for elem in self:
                yield self.state, elem

# %% ../nbs/10_callback.ipynb
def trackback(source: Iterable[Any], total: int|None|Type[EmptyT]=empty, context: Any=empty, 
        cbs:Sequence[Callback]=()) -> Iterator[tuple[AD[Any], Any]]:
    return CollBack(source, total, context, cbs=cbs).trackback()

# %% ../nbs/10_callback.ipynb
_T = TypeVar('_T')

def process_(
        iterable:Iterable[_T], /, 
        cbs: Callback|Sequence[Callback]=(), 
        slc:slice|None=None, 
        pred:Callable[[_T], bool]|None=None, 
        context:Any=empty,
        **kwargs  # FuncCB kwargs
    ) -> tuple[Callback,...]:
    "Process a subset `slc` of `iterable` filtered by `pred` with callbacks from `cbs` and `FuncCB` `kwargs`"
    _cbs = FC.tuplify(cbs) + ((FuncCB(**kwargs),) if kwargs else ())
    items = FC.L(iterable)[slc or slice(None)].filter(pred)  # type: ignore
    collections.deque(CollBack(items, context=context,cbs=_cbs), maxlen=0)
    return _cbs  # type: ignore



"""Fundamental Python utilities for data manipulation, type checking, and common operations"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_basic.ipynb.

# %% ../nbs/00_basic.ipynb 1
from __future__ import annotations

# %% auto 0
__all__ = ['empty', 'is_empty', 'AD', 'is_listy', 'is_listy_type', 'flatten', 'shorten', 'shortens', 'Runner', 'setattrs',
           'val_at', 'val_atpath', 'has_key', 'has_path', 'vals_atpath', 'vals_at', 'deep_in', 'pops_', 'pops_values_',
           'gets', 'update_', 'bundle_path', 'Kounter', 'simple_id', 'id_gen', 'WithCounterMeta']

# %% ../nbs/00_basic.ipynb
import importlib
import operator
import os
import pprint
import re
import sys
from binascii import hexlify
from inspect import Parameter
from pathlib import Path
from types import ModuleType
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Hashable
from typing import Iterable
from typing import Literal
from typing import Mapping
from typing import MutableMapping
from typing import Self
from typing import Sequence
from typing import Type
from typing import TypeAlias
from typing import TypeVar

import fastcore.all as FC


# %% ../nbs/00_basic.ipynb
class Empty(type): __repr__ = __str__ = lambda self: 'empty'
class EmptyT(metaclass=Empty):...

# _EMPTY: TypeAlias = Parameter.empty
# EmptyT = Type[_EMPTY]

# %% ../nbs/00_basic.ipynb
empty = EmptyT

def is_empty(x) -> bool: return x is empty

# %% ../nbs/00_basic.ipynb
_VT = TypeVar('_VT')
# from `fastcore` + generics
class AD(dict[str, _VT]):
    "`dict` subclass that also provides access to keys as attrs"
    def __getattr__(self, k:str) -> _VT: return self[k] if k in self else FC.stop(AttributeError(k))  # type: ignore
    def __setattr__(self, k, v:_VT): (self.__setitem__, super().__setattr__)[k[0]=='_'](k,v)
    def __dir__(self) -> Iterable[str]: return super().__dir__() + list(self.keys())  # type: ignore
    def _repr_markdown_(self): return f'```json\n{pprint.pformat(self, indent=2)}\n```'
    def copy(self) -> Self: return type(self)(**self)

# %% ../nbs/00_basic.ipynb
def is_listy(x):
    return isinstance(x, Iterable) and not isinstance(x, (bytes, str))

def is_listy_type(x):
    return issubclass(x, Iterable) and not issubclass(x, (bytes, str))

# %% ../nbs/00_basic.ipynb
def flatten(o):
    "Concatenate all collections and items as a generator"
    for item in o:
        if not is_listy(item): yield item; continue
        try: yield from flatten(item)
        except TypeError: yield item

# %% ../nbs/00_basic.ipynb
def shorten(x:Any, mode:Literal['l', 'r', 'c']='l', limit=40, trunc='…', empty='') -> str:
    if len(s := str(x)) > limit:
        l, m, r = (
            (empty, trunc, s[-limit:]) if mode == 'l' else 
            (s[:limit], trunc, empty) if mode == 'r' else 
            (s[:(limit//2)-1], f" {trunc} ", s[-(limit//2-1):])
        )
        s = f'{l}{m}{r}'
    return s

def shortens(xs:Iterable[Any], mode:Literal['l', 'r', 'c']='l', limit=40, trunc='…', empty=''):
    for x in xs: yield shorten(x, mode, limit, trunc, empty)


# %% ../nbs/00_basic.ipynb
_FuncItem: TypeAlias = Callable | Sequence['_FuncItem']

def Runner(*fns: _FuncItem) -> Callable:
    """Return a function that runs callables `fns` in sequence with same arguments. 
    Only side-effects, no composition."""
    _fns: tuple[Callable, ...] = tuple(FC.flatten(fns))  # type: ignore
    if not _fns: return FC.noop
    if len(_fns) == 1: return _fns[0]
    def _(*args, **kwargs) -> None:
        for f in _fns: f(*args, **kwargs)
    return _

# %% ../nbs/00_basic.ipynb
def setattrs(dest, src, flds=''):
    "Set `flds` or keys() or dir() attributes from `src` into `dest`"
    g = dict.get if isinstance(src, dict) else getattr
    s = operator.setitem if isinstance(dest, MutableMapping) else setattr
    if flds: flds = re.split(r",\s*", flds)
    elif isinstance(src, dict): flds = src.keys()
    else: flds = (_ for _ in dir(src) if _[0] != '_')
    for fld in flds: s(dest, fld, g(src, fld))

# %% ../nbs/00_basic.ipynb
def val_at(o, attr: str, default: Any=empty, sep='.'):
    "Traverse nested `o` looking for attributes/items specified in dot-separated `attr`."
    if not isinstance(attr, str): raise TypeError(f'{attr=!r} is not a string')
    try:
        for a in attr.split(sep):
            try: o = o[a]
            except (TypeError, KeyError):
                try: o = o[int(a)]
                except (IndexError, TypeError, KeyError, ValueError): o = getattr(o, a)
    except AttributeError as e:
        return default if default is not empty else FC.stop(e)  # type: ignore
    return o

def val_atpath(o, *path: Any,  default: Any=empty):
    "Traverse nested `o` looking for attributes/items specified in `path`."
    try:
        for a in path:
            try: o = o[a]
            except (IndexError, TypeError, KeyError):
                try: o = o[int(a)]
                except (IndexError, TypeError, KeyError, ValueError): o = getattr(o, str(a))
    except (AttributeError, TypeError) as e:
        return default if default is not empty else FC.stop(e)  # type: ignore
    return o

_NF = object()

def has_key(o, attr: str, sep='.') -> bool:
    "Return `True` if nested dot-separated `attr` exists."
    return val_at(o, attr, default=_NF, sep=sep) is not _NF

def has_path(o, *path: Any) -> bool:
    "Return `True` if nested `path` exists."
    return val_atpath(o, path, default=_NF) is not _NF

# %% ../nbs/00_basic.ipynb
def _vals_atpath(o, *path: Any, filter_empty=False) -> empty | tuple[empty | object, ...] | object:
    try: 
        idx = path.index('*'); pre, pos = path[:idx], path[idx+1:]
    except ValueError:
        return a if (a := val_atpath(o, *path, default=_NF)) is not _NF else empty
    a = val_atpath(o, *pre, default=_NF) if pre else o
    if not pos: return a
    if a is _NF: return empty
    try: 
        res = tuple(map(lambda x: _vals_atpath(x, *pos, filter_empty=filter_empty), a))  # type: ignore
        if all(x is empty for x in res): return empty
        return tuple(filter(lambda x: x is not empty, res)) if filter_empty else res
    except (AttributeError, TypeError) as e: return empty

def vals_atpath(o, *path: Any, filter_empty=False) -> tuple[Any, ...]:
    "Return nested values-- or empty|(empty, ...)-- at `path` with wildcards '*' from `d`."
    if '*' not in path: return () if (res := val_atpath(o, *path, default=_NF)) is _NF else (res,)
    res = _vals_atpath(o, *path, filter_empty=filter_empty)
    return () if res is empty else res  # type: ignore

def _vals_at(o, path:str, filter_empty=False) -> empty | tuple[empty | object, ...] | object:
    pre, wc, pos = str(path).partition('*')
    if not wc and not pos: return a if (a := val_at(o, pre, _NF)) is not _NF else empty
    a = val_at(o, pre.rstrip('.'), _NF) if pre else o
    if not pos: return a
    if a is _NF: return empty
    try: 
        res = tuple(map(lambda x: _vals_at(x, pos.lstrip('.'), filter_empty=filter_empty), a))  # type: ignore
        if all(x is empty for x in res): return empty
        return tuple(filter(lambda x: x is not empty, res)) if filter_empty else res
    except TypeError: return empty

def vals_at(o, path:str, filter_empty=False) -> tuple[Any, ...]:
    "Return nested values-- or empty|(empty, ...)-- at `path` with wildcards '*' from `o`."
    if '*' not in path: return () if (res := val_at(o, path, _NF)) is _NF else (res,)
    res = _vals_at(o, path, filter_empty=filter_empty)
    return () if res is empty else res  # type: ignore


# %% ../nbs/00_basic.ipynb
def deep_in(o:Mapping|Iterable, val):
    "return True if val is in nested collections"
    if isinstance(o, Mapping):
        for v in o.values():
            if v == val or (is_listy(v) and deep_in(v, val)): return True
        return False
    elif isinstance(o, Iterable):
        if val in o: return True
        return any(deep_in(v, val) for v in o if is_listy(v))
    raise ValueError(f"deep_in: o must be a Mapping or Iterable, got {type(o)}")

# %% ../nbs/00_basic.ipynb
def pops_(d: dict, *ks: Hashable) -> dict:
    "Pop existing `ks` items from `d` in-place into a dictionary."
    return {k:d.pop(k) for k in ks if k in d}

# %% ../nbs/00_basic.ipynb
def pops_values_(d: dict, *ks: Hashable) -> tuple:
    "Pop existing `ks` items from `d` in-place into a tuple of values or `Parameter.empty` for missing keys."
    return tuple(d.pop(k, Parameter.empty) for k in ks)

# %% ../nbs/00_basic.ipynb
def gets(d: Mapping, *ks: Hashable):
    "Fetches `ks` values, or `Parameter.empty` for missing keys, from `d` into a tuple."
    return tuple(d.get(k, Parameter.empty) for k in ks)  # type: ignore

# %% ../nbs/00_basic.ipynb
def update_(dest=None, /, empty_value=None, **kwargs) -> Any:
    "Update `dest` in-place with `kwargs` whose values aren't `empty_value`"
    dest = dest if dest is not None else {}
    f = operator.setitem if isinstance(dest, MutableMapping) else setattr
    for k, v in filter(lambda x: x[1]!=empty_value, kwargs.items()): f(dest, k, v)
    return dest

# %% ../nbs/00_basic.ipynb
def _get_globals(mod: str):
    if hasattr(sys, '_getframe'):
        glb = sys._getframe(2).f_globals
    else:
        glb = sys.modules[mod].__dict__
    return glb

# %% ../nbs/00_basic.ipynb
def bundle_path(mod:str|ModuleType):
    "Return the path to the module's directory or current directory."
    if isinstance(mod, str): mod = importlib.import_module(mod)
    return Path(fn).parent if (fn := getattr(mod, '__file__', None)) else Path()

# %% ../nbs/00_basic.ipynb
class Kounter:
    def __init__(self): self.d = DefaultDict(int)
    def __call__(self, k): d = self.d; d[k] += 1; return self.d[k]

# %% ../nbs/00_basic.ipynb
def simple_id():
    return 'b'+hexlify(os.urandom(16), '-', 4).decode('ascii')

def id_gen():
    kntr = Kounter()
    def _(o:Any=None): 
        if o is None: return simple_id()
        # return f"{type(o).__name__}_{hash(o) if isinstance(o, Hashable) else kntr(type(o).__name__)}"
        return f"{type(o).__name__}_{kntr(type(o).__name__)}"
    return _

# %% ../nbs/00_basic.ipynb
class WithCounterMeta(FC.FixSigMeta):
    "Adds a `_cnt_` attribute to its classes and increments it for each new instance."
    _cnt_: int
    def __call__(cls, *args, **kwargs):
        res = super().__call__(*args, **kwargs)
        res._cnt_ = cls._cnt_
        cls._cnt_ += 1
        return res
    def __new__(cls, name, bases, dict):
        res = super().__new__(cls, name, bases, dict)
        res._cnt_ = 0
        return res



"""Global console configuration for consistent output across projects"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_project.ipynb.

# %% ../nbs/00_project.ipynb 1
from __future__ import annotations

# %% auto 0
__all__ = ['get_console', 'setup_console']

# %% ../nbs/00_project.ipynb
import pathlib
from typing import Callable

import fastcore.all as FC
from rich.console import Console


# %% ../nbs/00_project.ipynb
# _console = Console(width=107, tab_size= 4, theme=dark_theme, force_jupyter=True if FC.IN_NOTEBOOK else None)
_CONSOLE: Console = None  # type: ignore

def get_theme(theme:str | object | None):
    if isinstance(theme, str):
        try:
            from rich_theme_manager import ThemeManager  # type: ignore
            theme_dir = pathlib.Path("~/.rich_theme_manager/themes").expanduser()
            theme_manager = ThemeManager(theme_dir=theme_dir.absolute().as_posix())
            try: theme = theme_manager.get(theme)
            except ValueError: theme = None
        except ImportError:
            theme = None
    # return theme or Theme('default', styles={"repr.tag_contents": "white"})
    return theme


# %% ../nbs/00_project.ipynb
def get_console() -> Console:
    "Globally configured console"
    global _CONSOLE
    return _CONSOLE

def setup_console(
        width: int = 140,
        theme: str | object | None = None,
        clear=False,
        rule=None
) -> tuple[Console, Callable[..., None]]:
    global _CONSOLE
    theme = get_theme(theme)
    params = dict(width=width, tab_size=4, theme=theme, force_jupyter=True if FC.IN_IPYTHON else None)
    _CONSOLE = Console(**params)  # type: ignore
    if clear: _CONSOLE.clear()
    if rule:_CONSOLE.rule(rule)
    return _CONSOLE, _CONSOLE.print

setup_console();
